"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.evaluateExpression = void 0;
var t = __importStar(require("@babel/types"));
var traverse_1 = __importDefault(require("@babel/traverse"));
var ast_1 = require("./ast");
var createResultPair = function (value, meta) { return ({
    value: value,
    meta: meta,
}); };
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there is an identifier called `color` that is set somewhere as `const color = 'blue'`,
 * passing the `color` identifier to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param state Babel state - should house options and meta data used during the transformation.
 */
var traverseIdentifier = function (expression, meta) {
    var _a;
    var value = undefined;
    var updatedMeta = meta;
    var resolvedBinding = ast_1.resolveBindingNode(expression.name, updatedMeta);
    if (resolvedBinding && resolvedBinding.constant) {
        // We recursively call get interpolation until it not longer returns an identifier or member expression
        (_a = exports.evaluateExpression(resolvedBinding.node, resolvedBinding.meta), value = _a.value, updatedMeta = _a.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will evaluate object values recursively and return the actual value along with updated metadata.
 *
 * E.g: If there is an object expression `{ x: () => 10 }`, it will evaluate and
 * return `value` as `10`.
 * @param expression Expression we want to interrogate.
 * @param accessPath An array of nested object keys
 * @param meta Meta data used during the transformation.
 */
var evaluateObjectExpression = function (expression, accessPath, meta) {
    var _a;
    var value = expression;
    var updatedMeta = meta;
    if (t.isObjectExpression(expression)) {
        var objectValue = ast_1.getValueFromObjectExpression(expression, accessPath);
        (_a = exports.evaluateExpression(objectValue, updatedMeta), value = _a.value, updatedMeta = _a.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there is a member expression called `colors().primary` that has identifier `colors` which
 * is set somewhere as `const colors = () => ({ primary: 'blue' })`,
 * passing the `colors` identifier to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param accessPath An array of nested object keys
 * @param meta Meta data used during the transformation.
 */
var evaluateCallExpressionBindingMemberExpression = function (expression, accessPath, meta) {
    var _a, _b;
    var value = expression;
    var updatedMeta = meta;
    if (t.isFunction(expression)) {
        (_a = exports.evaluateExpression(expression, meta), value = _a.value, updatedMeta = _a.meta);
        (_b = evaluateObjectExpression(value, accessPath, updatedMeta), value = _b.value, updatedMeta = _b.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g:
 * 1. If there is a member expression called `colors.primary` that has identifier `colors` which
 * is set somewhere as `const colors = { primary: 'blue' }`,
 * passing the `colors` identifier to this function would return `'blue'`.
 *
 * 2. If there is a member expression called `colors.primary` that has identifier `colors` which
 * is set somewhere as `const colors = colorMixin();` calling another identifier
 * `const colorMixin = () => ({ primary: 'blue' })`, passing the `colors` identifier
 * to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param accessPath An array of nested object keys
 * @param meta Meta data used during the transformation.
 */
var evaluateIdentifierBindingMemberExpression = function (expression, accessPath, meta) {
    var _a, _b, _c;
    var value = expression;
    var updatedMeta = meta;
    if (t.isObjectExpression(expression)) {
        (_a = evaluateObjectExpression(expression, accessPath, meta), value = _a.value, updatedMeta = _a.meta);
    }
    else if (t.isCallExpression(expression)) {
        (_b = exports.evaluateExpression(expression, meta), value = _b.value, updatedMeta = _b.meta);
        (_c = evaluateObjectExpression(value, accessPath, updatedMeta), value = _c.value, updatedMeta = _c.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there is a member expression called `colors.primary` that has identifier `color` which
 * is set somewhere as `const colors = { primary: 'blue' }`,
 * passing the `colors` identifier to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param state Babel state - should house options and meta data used during the transformation.
 */
var traverseMemberExpression = function (expression, meta) {
    var _a, _b;
    var value = undefined;
    var updatedMeta = meta;
    var _c = ast_1.getMemberExpressionMeta(expression), accessPath = _c.accessPath, bindingIdentifier = _c.bindingIdentifier, originalBindingType = _c.originalBindingType;
    if (bindingIdentifier) {
        var resolvedBinding = ast_1.resolveBindingNode(bindingIdentifier.name, updatedMeta);
        if (resolvedBinding && resolvedBinding.constant && t.isExpression(resolvedBinding.node)) {
            if (originalBindingType === 'Identifier') {
                (_a = evaluateIdentifierBindingMemberExpression(resolvedBinding.node, accessPath, resolvedBinding.meta), value = _a.value, updatedMeta = _a.meta);
            }
            else if (originalBindingType === 'CallExpression') {
                (_b = evaluateCallExpressionBindingMemberExpression(resolvedBinding.node, accessPath, resolvedBinding.meta), value = _b.value, updatedMeta = _b.meta);
            }
        }
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there was a function called `size` that is set somewhere as
 * `const size = () => 10` or `const size = function() { return 10; }` or `function size() { return 10; }`,
 * passing the `size` identifier to this function would return `10` (it will recursively evaluate).
 *
 * @param expression Expression we want to interrogate.
 * @param state Babel state - should house options and meta data used during the transformation.
 */
var traverseFunction = function (expression, meta) {
    var _a;
    var value = undefined;
    var updatedMeta = meta;
    if (t.isBlockStatement(expression.body)) {
        traverse_1.default(expression.body, {
            noScope: true,
            ReturnStatement: function (path) {
                var _a;
                var argument = path.node.argument;
                if (argument) {
                    (_a = exports.evaluateExpression(argument, meta), value = _a.value, updatedMeta = _a.meta);
                }
                path.stop();
            },
        });
    }
    else {
        (_a = exports.evaluateExpression(expression.body, meta), value = _a.value, updatedMeta = _a.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will find the function node for the call expression and wrap an IIFE around it (to avoid name collision)
 * and move all the parameters mapped to passed arguments in the IIFE's scope (own scope in this case).
 * It will also set own scope path so that when we recursively evaluate any node,
 * we will look for its binding in own scope first, then parent scope.
 *
 * @param expression Expression we want to interrogate.
 * @param meta Meta data used during the transformation.
 */
var traverseCallExpression = function (expression, meta) {
    var _a;
    var callee = expression.callee;
    var value = undefined;
    var updatedMeta = meta;
    /*
      Basically flow is as follows:
      1. Get the calling function node. It can be either identifier `func('arg')` or a
         member expression `x.func('arg')`;
      2. Save the reference of function node for example `func`.
      3. Pull out its parameters.
      4. Evaluate the args of calling function `func('arg')` by traversing recursively.
      5. Loop through params, and map the args with params.
      6. Pull the params and create an IIFE around the calling function `CallExpression` which creates
         a new scope around calling function and isolates everything (We add bindings at this step by pushing into its scope).
      7. Add that IIFE node path to `ownPath`, and check only for own binding in `resolveBindingNode` function
         so that only isolated params are evaluated.
      8. In `resolveBindingNode`, if `ownPath` is not set (module traversal case or any other case), it will pick things from `parentPath`.
    */
    if (t.isExpression(callee)) {
        var functionNode = null;
        // Get func node either from `Identifier` i.e. `func('arg')` or `MemberExpression` i.e. `x.func('arg')`.
        // Right now we are only supported these 2 flavors. If we have complex case like `func('arg').fn().variable`,
        // it will not get evaluated.
        if (t.isIdentifier(callee)) {
            var resolvedBinding = ast_1.resolveBindingNode(callee.name, updatedMeta);
            if (resolvedBinding && resolvedBinding.constant) {
                functionNode = resolvedBinding.node;
            }
        }
        else if (t.isMemberExpression(callee) && t.isIdentifier(callee.property)) {
            var _b = ast_1.getMemberExpressionMeta(callee), accessPath = _b.accessPath, bindingIdentifier = _b.bindingIdentifier;
            if (bindingIdentifier) {
                var resolvedBinding = ast_1.resolveBindingNode(bindingIdentifier.name, updatedMeta);
                if (resolvedBinding && resolvedBinding.constant) {
                    if (t.isObjectExpression(resolvedBinding.node)) {
                        functionNode = ast_1.getValueFromObjectExpression(resolvedBinding.node, accessPath);
                    }
                }
            }
        }
        // Check if it is resolved
        if (functionNode && t.isFunction(functionNode)) {
            // Pull its parameters
            var params = functionNode.params;
            // Evaluate the passed args recursively
            var evaluatedArguments_1 = expression.arguments.map(function (argument) { return exports.evaluateExpression(argument, updatedMeta).value; });
            // Get path of the call expression `func('arg')` or `x.func('arg')`
            var expressionPath = ast_1.getPathOfNode(expression, updatedMeta.parentPath);
            // Create an IIFE around it and replace its path. So `func('arg')` will become `(() => func('arg'))()`
            // and `x.func('arg')` will become `(() => x.func('arg'))()`.
            // `wrappingNodePath` is the path above.
            var wrappingNodePath = expressionPath.replaceWith(ast_1.wrapNodeInIIFE(expression))[0];
            // Get arrowFunctionExpressionPath, which was created using the IIFE
            var arrowFunctionExpressionPath_1 = ast_1.getPathOfNode(wrappingNodePath.node.callee, wrappingNodePath);
            // Loop through the parameters. Right now only identifier `param` and object pattern `{ param }` or `{ param: p }`
            // are supported.
            params
                .filter(function (param) { return t.isIdentifier(param) || t.isObjectPattern(param); })
                .forEach(function (param, index) {
                var evaluatedArgument = evaluatedArguments_1[index];
                // Push evaluated args and params in the IIFE's scope by created a local variable
                // `const param = 'evaluated arg value'`
                arrowFunctionExpressionPath_1.scope.push({
                    id: param,
                    init: evaluatedArgument,
                    kind: 'const',
                });
            });
            // Set the `ownPath` which `resolveBindingNode` will use to check own binding for the
            // local variables we created above.
            updatedMeta.ownPath = arrowFunctionExpressionPath_1;
        }
        (_a = exports.evaluateExpression(callee, updatedMeta), value = _a.value, updatedMeta = _a.meta);
    }
    return createResultPair(value, updatedMeta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 * If the expression is an identifier node (a variable) and a constant,
 * it will return the variable reference.
 *
 * E.g: If there was a identifier called `color` that is set somewhere as `const color = 'blue'`,
 * passing the `color` identifier to this function would return `'blue'`.
 *
 * This behaviour is the same for const string & numeric literals,
 * and object expressions.
 *
 * @param expression Expression we want to interrogate.
 * @param state Babel state - should house options and meta data used during the transformation.
 */
var evaluateExpression = function (expression, meta) {
    var _a, _b, _c, _d;
    var value = undefined;
    var updatedMeta = meta;
    if (t.isIdentifier(expression)) {
        (_a = traverseIdentifier(expression, updatedMeta), value = _a.value, updatedMeta = _a.meta);
    }
    else if (t.isMemberExpression(expression)) {
        (_b = traverseMemberExpression(expression, updatedMeta), value = _b.value, updatedMeta = _b.meta);
    }
    else if (t.isFunction(expression)) {
        (_c = traverseFunction(expression, updatedMeta), value = _c.value, updatedMeta = _c.meta);
    }
    else if (t.isCallExpression(expression)) {
        (_d = traverseCallExpression(expression, updatedMeta), value = _d.value, updatedMeta = _d.meta);
    }
    if (t.isStringLiteral(value) || t.isNumericLiteral(value) || t.isObjectExpression(value)) {
        return createResultPair(value, updatedMeta);
    }
    // --------------
    // NOTE: We are recursively calling evaluateExpression() which is then going to try and evaluate it
    // multiple times. This may or may not be a performance problem - when looking for quick wins perhaps
    // there is something we could do better here.
    // --------------
    if (value) {
        return createResultPair(ast_1.tryEvaluateExpression(value, updatedMeta, expression), updatedMeta);
    }
    return createResultPair(ast_1.tryEvaluateExpression(expression, updatedMeta), updatedMeta);
};
exports.evaluateExpression = evaluateExpression;
//# sourceMappingURL=evaluate-expression.js.map